package data;

import java.util.*;
import java.awt.*;

public class WorldMap {

    private Display display = null;
    private int width;
    private int height;

    private GridItem[][] worldMap;
    private Vector homeBases = new Vector(); // contains GridItem
    private Hashtable packageOnMap = new Hashtable();
    private Hashtable packageCarried = new Hashtable();
    
    public WorldMap(Display d, int w, int h) {
        display = d;
        init(w, h);
    }
    
    public WorldMap(int w, int h) {
        init(w, h);
    }

    private void init(int w, int h) {
        width = w;
        height = h;
        System.out.println("Create Map " + width + "x" + height);
        worldMap = new GridItem[width][height];
    }

    public int getWidth() { return width; }
    public int getHeight() { return height; }

    public int totalPackages() {
        System.out.println("ON MAP  " + packageOnMap.size());
        System.out.println("CARRIED " + packageCarried.size());
        return packageOnMap.size() + packageCarried.size();
    }
    
    public void addItem(int col, int row, char c) throws Exception {
        // translate into SW coordinates
        int x = col - 1;
        int y = height - row;
        
        GridItem gi = new GridItem(c, col, row);
        worldMap[x][y] = gi;
        if (gi.type == GridItem.TYPE_HOME) {
            homeBases.addElement(gi);
        }
    }

    public Robot addRobot(String id, int col, int row) {
        // translate into SW coordinates
        int x = col - 1;
        int y = height - row;
        
        GridItem item = worldMap[x][y];

        Robot robot = new Robot(id, Color.green, col, row);
        robot.me = true;
        item.addRobot(robot);
        if (display != null) display.updateGrid(col, row);
        return robot;
    }
    
    public void addEnemyRobot(String id, int col, int row) {
        // translate into SW coordinates
        int x = col - 1;
        int y = height - row;
        
        GridItem item = worldMap[x][y];

        Robot robot = new Robot(id, Color.red, col, row);
        item.addRobot(robot);
        if (display != null) display.updateGrid(col, row);
    }
    
    public void addPackage(PackageItem item, int col, int row) {
        // translate into SW coordinates
        int x = col - 1;
        int y = height - row;
        
        packageOnMap.put(item.id, item);
        worldMap[x][y].addPackage(item);
    }

    public String getPackageInfo(String robotID) {
        GridItem item = findRobotGrid(robotID);
        if (item != null) {
            return item.getPackageString();
        }
        return "";
    }

    public GridItem findRobotGrid(String robotID) {
        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {
                GridItem item = worldMap[x][y];
                if (item.containsRobot(robotID)) {
                    return item;
                }
            }
        }
        return null;
    }

    public GridItem findUnexploredBase() {
        GridItem exploreHome = null;
        for (Enumeration h = homeBases.elements(); h.hasMoreElements(); ) {
            GridItem gi = (GridItem) h.nextElement();
            if (!gi.explored) {
                exploreHome = gi;
                break;
            }
        }
        return exploreHome;
    }
        
        
    public boolean allBasesOccupied() {
        boolean empty = true;

        for (Enumeration b = homeBases.elements(); b.hasMoreElements(); ) {
            GridItem item = (GridItem) b.nextElement();
            if (item.containsRobot("")) {
                empty = false;
            }
        }
        
        return empty;
    }
    
    public GridItem findEmptyBase() {
        GridItem home = null;
        for (Enumeration h = homeBases.elements(); h.hasMoreElements(); ) {
            GridItem item = (GridItem) h.nextElement();
            if (item.containsRobot("")) {
                home = item;
                break;
            }
        }
        return home;
    }

    public GridItem getGridItem(int col, int row) {
        return worldMap[col-1][height - row];
    }

    public String getRobotLocations() {
        StringBuffer buf = new StringBuffer();
        for (int y = 0; y < height; y++) {
            for (int x = 0; x < width; x++) {
                GridItem item = worldMap[x][y];
                if (item.hasRobots()) {
                    Iterator iter = item.robots.values().iterator();
                    while (iter.hasNext()) {
                        Robot robot = (Robot) iter.next();
                        buf.append("#" + robot + " X " + (x + 1) +
                                   " Y " + (height - y) + " ");
                    }
                }
            }
        }
        return buf.toString();
    }

    public String getRow(int row, boolean simple) {
        // translate into SW coordinates
        int y = height - row;
        
        StringBuffer buf = new StringBuffer();
        for (int x = 0; x < width; x++) {
            buf.append(worldMap[x][y].toString(simple));
        }
        return buf.toString();
    }

    public String pick(String robotID, String ids) {
        String result = "#" + robotID;
        GridItem item = findRobotGrid(robotID);

        StringTokenizer st = new StringTokenizer(ids, " ");
        while (st.hasMoreTokens()) {
            String pId = st.nextToken();
            PackageItem pItem = (PackageItem) packageOnMap.get(pId);
            item.pickPackage(pId);
            packageOnMap.remove(pId);
            packageCarried.put(pId, pItem);
            result += " P " + pId;
        }

        return result;
        
    }
    
    public String drop(String robotID, String ids) {
        String result = "#" + robotID;
        GridItem gridItem = findRobotGrid(robotID);

        StringTokenizer st = new StringTokenizer(ids, " ");
        while (st.hasMoreTokens()) {
            String pId = st.nextToken();
            PackageItem pItem = (PackageItem) packageCarried.get(pId);
            if (!gridItem.dropPackage(pItem)) {
                // they dropped it back on the map
                packageOnMap.put(pId, pItem);
            }
            packageCarried.remove(pId);
            result += " D " + pItem.id;
        }

        return result;
    }
    
    public String move(Robot robot, String dir) {
        String result = "";
        
        int x = robot.col - 1;
        int y = height - robot.row;

        GridItem item = getGridItem(robot.col, robot.row);
        item.removeRobot(robot);
        
        if (dir.equals("N")) {
            item = worldMap[x][y-1];
        } else
        if (dir.equals("S")) {
            item = worldMap[x][y+1];
        } else
        if (dir.equals("E")) {
            item = worldMap[x+1][y];
        } else
        if (dir.equals("W")) {
            item = worldMap[x-1][y];
        }
        item.addRobot(robot);
        
        return "#" + robot.robotID + " " + dir + " ";
    }

    public void explored(int col, int row) {
        getGridItem(col, row).explored = true;
    }

    public GridItem getUnexploredGridItem(int col, int row) {        
        for (int d = 1; d < Math.max(width, height); d++) {
            for (int c = col - d; c < (col + d); c++) {
                for (int r = row + d; r > (row - d); r--) {
                    if ((c > 0) && (r > 0) && (c <= width) && (r <= height)) {
                        GridItem item = getGridItem(c,r);
                        if (!item.explored) {
                            return item;
                        }
                    }
                }
            }
        }
        return null;
    }
    
    public String toString() {
        StringBuffer buf = new StringBuffer();
        for (int row = height; row > 0; row--) {
            buf.append(getRow(row, false) + "\n");
        }
        return buf.toString();
    }
}
